from datetime import datetime, timedelta
import os
import sqlite3

from rich import print

from .config import DB_PATH, DB_VERSION, SQL_DIR
from .singleton_meta import SingletonMeta


class Database(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self._connect()
        if self._db_exists():
            self._update_db()
        else:
            self._create_db()
        self.cur.execute("PRAGMA foreign_keys=ON")

    def _connect(self) -> None:
        self.conn = sqlite3.connect(DB_PATH)
        self.cur = self.conn.cursor()

    def _db_exists(self) -> bool:
        self.cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
        tables = {row[0] for row in self.cur.fetchall()}
        return True if tables == {"sessions", "tasks", "projects"} else False

    def _create_db(self) -> None:
        with open(os.path.join(SQL_DIR, "createdb.sql"), "r") as file:
            sql = file.read()
        self.cur.executescript(sql)
        self.conn.commit()

    def _update_db(self) -> None:
        self.cur.execute("PRAGMA user_version")
        user_version = self.cur.fetchone()[0]
        if user_version < DB_VERSION:
            try:
                self._migrate(user_version)
            except FileNotFoundError:
                print(
                    "[red]Whoops. Seems like wrong db version.[/]\n"
                    "[green]Contact developer.\n[/]"
                    "[blue]https://github.com/markbragin/mrtracker"
                )
                exit(2)

    def _migrate(self, old_version: int) -> None:
        sql_script = f"migrate_from_{old_version}.sql"
        with open(os.path.join(SQL_DIR, sql_script), "r") as file:
            sql = file.read()
        self.cur.executescript(sql)
        self.conn.commit()

    def fetch_tasks(self) -> list[tuple]:
        self.cur.execute(
            "SELECT t.id, p.id, t.name, SUM(s.duration), t.tags "
            "FROM tasks t "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "LEFT JOIN sessions s "
            "ON t.id = s.task_id "
            "GROUP BY t.id "
            "ORDER BY t.id"
        )
        return self.cur.fetchall()

    def fetch_projects(self) -> list[tuple]:
        self.cur.execute(
            "SELECT id, NULL, name, NULL, NULL FROM projects p ORDER BY p.id"
        )
        return self.cur.fetchall()

    def fetch_tasks_today(self) -> list[tuple]:
        self.cur.execute(
            "SELECT t.name, SUM(s.duration) sum, p.name "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date = date('now', 'localtime') "
            "GROUP BY s.task_id "
            "ORDER BY sum DESC"
        )
        return self.cur.fetchall()

    def fetch_tasks_week(self) -> list[tuple]:
        now = datetime.now()
        week_ago = now - timedelta(days=7)
        self.cur.execute(
            "SELECT t.name, SUM(s.duration) sum, p.name "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date BETWEEN (?) AND (?) "
            "GROUP BY s.task_id "
            "ORDER BY sum DESC",
            (week_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_tasks_month(self) -> list[tuple]:
        now = datetime.now()
        month_ago = now - timedelta(days=30)
        self.cur.execute(
            "SELECT t.name, SUM(s.duration) sum, p.name "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date BETWEEN (?) AND (?) "
            "GROUP BY s.task_id "
            "ORDER BY sum DESC",
            (month_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_projects_today(self) -> list[tuple]:
        self.cur.execute(
            "SELECT p.name, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date = date('now', 'localtime') "
            "GROUP BY p.id "
            "ORDER BY sum DESC"
        )
        return self.cur.fetchall()

    def fetch_projects_week(self) -> list[tuple]:
        now = datetime.now()
        week_ago = now - timedelta(days=7)
        self.cur.execute(
            "SELECT p.name, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date BETWEEN (?) AND (?) "
            "GROUP BY p.id "
            "ORDER BY sum DESC",
            (week_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_projects_month(self) -> list[tuple]:
        now = datetime.now()
        month_ago = now - timedelta(days=30)
        self.cur.execute(
            "SELECT p.name, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
            "WHERE date BETWEEN (?) AND (?) "
            "GROUP BY p.id "
            "ORDER BY sum DESC",
            (month_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_tags_today(self) -> list[tuple]:
        self.cur.execute(
            "SELECT t.tags, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "WHERE t.tags IS NOT NULL AND date = date('now', 'localtime') "
            "GROUP BY t.tags "
            "ORDER BY sum DESC"
        )
        return self.cur.fetchall()

    def fetch_tags_week(self) -> list[tuple]:
        now = datetime.now()
        week_ago = now - timedelta(days=7)
        self.cur.execute(
            "SELECT t.tags, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "WHERE t.tags IS NOT NULL AND date BETWEEN (?) AND (?) "
            "GROUP BY t.tags "
            "ORDER BY sum DESC",
            (week_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_tags_month(self) -> list[tuple]:
        now = datetime.now()
        month_ago = now - timedelta(days=30)
        self.cur.execute(
            "SELECT t.tags, SUM(s.duration) sum "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON t.id = s.task_id "
            "WHERE t.tags IS NOT NULL AND date BETWEEN (?) AND (?) "
            "GROUP BY t.tags "
            "ORDER BY sum DESC",
            (month_ago.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d")),
        )
        return self.cur.fetchall()

    def fetch_for_csv(self) -> list[tuple]:
        """fetches project|task|tags|date|start_time|end_time|duration(str)"""
        self.cur.execute(
            "SELECT p.name, t.name, t.tags, s.date, "
            "s.start_time, s.end_time, time(s.duration, 'unixepoch') "
            "FROM sessions s "
            "LEFT JOIN tasks t "
            "ON s.task_id = t.id "
            "LEFT JOIN projects p "
            "ON t.project_id = p.id "
        )
        return self.cur.fetchall()

    def add_session(
        self,
        task_id: int,
        date: str,
        start_time: str,
        end_time: str,
    ) -> None:
        self.cur.execute(
            "INSERT INTO sessions (task_id, date, start_time, end_time) "
            "VALUES (?, ?, ?, ?)",
            (task_id, date, start_time, end_time),
        )
        self.conn.commit()

    def add_project(self, name: str) -> None:
        self.cur.execute("INSERT INTO projects (name) VALUES (?)", (name,))
        self.conn.commit()

    def add_task(self, name: str, project_id: int) -> None:
        self.cur.execute(
            "INSERT INTO tasks (name, project_id) VALUES (?, ?)",
            (name, project_id),
        )
        self.conn.commit()

    def delete_project(self, project_id: int) -> None:
        self.cur.execute("DELETE FROM projects WHERE id=(?)", (project_id,))
        self.conn.commit()

    def delete_task(self, task_id: int) -> None:
        self.cur.execute("DELETE FROM tasks WHERE id=(?)", (task_id,))
        self.conn.commit()

    def rename_project(self, project_id: int, new_name: str) -> None:
        self.cur.execute(
            "UPDATE projects SET name = (?) WHERE id=(?)",
            (new_name, project_id),
        )
        self.conn.commit()

    def rename_task(self, task_id: int, new_name: str) -> None:
        self.cur.execute(
            "UPDATE tasks SET name = (?) WHERE id=(?)",
            (new_name, task_id),
        )
        self.conn.commit()

    def update_tags(self, task_ids: list[int], tag: str | None) -> None:
        placeholders = ", ".join("?" for _ in task_ids)
        self.cur.execute(
            "UPDATE tasks SET tags=(?) WHERE id in (%s)" % placeholders,
            (tag, *task_ids),
        )
        self.conn.commit()

    def change_project(self, task_id: int, new_project_id: int) -> None:
        self.cur.execute(
            "UPDATE tasks SET project_id = (?) WHERE id=(?)",
            (new_project_id, task_id),
        )
        self.conn.commit()

    def swap_projects(self, id1: int, id2: int) -> None:
        self.cur.execute("UPDATE projects SET id=-1 WHERE id=(?)", (id1,))
        self.cur.execute("UPDATE projects SET id=(?) WHERE id=(?)", (id1, id2))
        self.cur.execute("UPDATE projects SET id=(?) WHERE id=-1", (id2,))
        self.conn.commit()

    def swap_tasks(self, id1: int, id2: int) -> None:
        self.cur.execute("UPDATE tasks SET id=-1 WHERE id=(?)", (id1,))
        self.cur.execute("UPDATE tasks SET id=(?) WHERE id=(?)", (id1, id2))
        self.cur.execute("UPDATE tasks SET id=(?) WHERE id=-1", (id2,))
        self.conn.commit()

    def delete_sessions_by_task_ids(self, task_ids: list[int]) -> None:
        placeholders = ", ".join("?" for _ in task_ids)
        self.cur.execute(
            "DELETE FROM sessions WHERE task_id in (%s)" % placeholders,
            task_ids,
        )
        self.conn.commit()

    def get_next_task_id(self) -> int:
        self.cur.execute("SELECT MAX(id) FROM tasks")
        id = self.cur.fetchone()[0]
        return id + 1 if id else 1

    def get_next_project_id(self) -> int:
        self.cur.execute("SELECT MAX(id) FROM projects")
        id = self.cur.fetchone()[0]
        return id + 1 if id else 1



from rich.console import RenderableType
from rich.text import Text
from textual.app import events
from textual.reactive import Reactive
from textual.widget import Widget


class TextInput(Widget):

    _content: Reactive[str] = Reactive("")
    _cursor_pos: Reactive[int] = Reactive(0)
    _has_focus: Reactive[bool] = Reactive(False)
    _width: int = 0
    placeholder: str = ""

    def __init__(self, name: str | None = "TextInput") -> None:
        super().__init__(name=name)

    def on_focus(self) -> None:
        self._has_focus = True

    def on_blur(self) -> None:
        self._has_focus = False

    async def _unfocus(self) -> None:
        await self.app.set_focus(None)

    def clear_content(self) -> None:
        self._content = ""
        self._width = 0
        self._cursor_pos = 0

    @property
    def content(self) -> str:
        return self._content

    @content.setter
    def content(self, new_content: str) -> None:
        self._content = new_content
        self._width = len(new_content)
        self._cursor_pos = self._width

    @property
    def has_focus(self) -> bool:
        return self._has_focus

    def on_key(self, event: events.Key) -> None:
        if event.key == "ctrl+h":
            self._on_backspace()
        elif event.key == "delete":
            self._on_delete()
        elif event.key == "left":
            self._cursor_left()
        elif event.key == "escape":
            return
        elif event.key == "right":
            self._cursor_right()
        elif event.key.isprintable() and len(event.key) == 1:
            self._insert_symbol(event.key)
        else:
            return
        event.stop()

    def _insert_symbol(self, symbol: str) -> None:
        self._content = (
            self._content[: self._cursor_pos]
            + symbol
            + self._content[self._cursor_pos :]
        )
        self._width += 1
        self._cursor_right()

    def _on_backspace(self) -> None:
        if self._cursor_pos > 0:
            self._content = (
                self._content[: self._cursor_pos - 1]
                + self._content[self._cursor_pos :]
            )
            self._width -= 1
            self._cursor_left()

    def _on_delete(self) -> None:
        if self._cursor_pos == self._width - 1:
            self._content = self._content[: self._cursor_pos]
        elif self._cursor_pos < self._width - 1:
            self._content = (
                self._content[: self._cursor_pos]
                + self._content[self._cursor_pos + 1 :]
            )
        else:
            return
        self._width -= 1

    def _is_valid_pos(self, pos: int) -> bool:
        return 0 <= pos <= self._width

    def _cursor_left(self) -> None:
        if self._is_valid_pos(self._cursor_pos - 1):
            self._cursor_pos -= 1

    def _cursor_right(self) -> None:
        if self._is_valid_pos(self._cursor_pos + 1):
            self._cursor_pos += 1

    def render(self) -> RenderableType:
        if self.has_focus:
            return self._render_with_cursor()
        else:
            return self._content

    def _render_with_cursor(self) -> Text:
        if self._cursor_pos == self._width:
            visible = Text.assemble(
                self._content[: self._cursor_pos], Text(" ", style="reverse")
            )
        else:
            visible = Text.assemble(
                self._content[: self._cursor_pos],
                Text(f"{self._content[self._cursor_pos]}", style="reverse"),
                self._content[self._cursor_pos + 1 :],
            )

        return visible

    def _render_paceholder(self) -> RenderableType:
        return self.placeholder



from rich.padding import PaddingDimensions
from rich.text import TextType
from textual.app import events
from textual.widgets import NodeID, TreeControl, TreeNode


class NestedList(TreeControl):
    def __init__(
        self,
        label: TextType,
        data,
        name: str | None = None,
        padding: PaddingDimensions = (1, 1),
    ) -> None:
        super().__init__(label, data, name=name, padding=padding)
        self.show_cursor = True

    async def add_child(self, label: TextType, data) -> None:
        await self.add(self.cursor, label=label, data=data)
        await self.nodes[self.cursor].expand()
        self._cur_to_latest_child()

    async def add_sibling(self, label: TextType, data) -> None:
        node = self.nodes[self.cursor]
        if node is self.root:
            return
        self._cur_to_parent()
        await self.add_child(label, data)

    async def add_root_child(self, label: TextType, data) -> None:
        await self._cur_to_root()
        await self.add_child(label, data)

    async def remove_node(self, node: TreeNode | None = None) -> None:
        node = node if node else self.nodes[self.cursor]

        if node is self.root:
            return

        for nd in node.children:
            await self.remove_node(nd)

        if node.id == self.cursor:
            await self.cursor_up()

        node.parent.tree.children.remove(node.tree)
        node.parent.children.remove(node)
        del self.nodes[node.id]
        self.id = max(self.nodes.keys())

    def _cur_to_latest_child(self) -> None:
        self.cursor = self.id

    def _cur_to_parent(self) -> None:
        parent = self.nodes[self.cursor].parent
        self.cursor = parent.id if parent else self.cursor

    def _cur_up_to_root_child(self) -> None:
        while self.nodes[self.cursor].parent is not self.root:
            self._cur_to_parent()

    async def _cur_to_next_sibling(self) -> None:
        sibling = self.nodes[self.cursor].next_sibling
        if sibling:
            while self.cursor != sibling.id:
                await self.cursor_down()

    async def _cur_to_root(self) -> None:
        self.cursor = NodeID(0)
        self.cursor_line = 0

    async def toggle_folding(self) -> None:
        await self.nodes[self.cursor].toggle()

    async def key_down(self, event: events.Key) -> None:
        pass

    async def key_up(self, event: events.Key) -> None:
        pass

    async def key_enter(self, event: events.Key) -> None:
        pass



from datetime import datetime
import json
import os
import shutil

from platformdirs import user_config_dir, user_data_dir


APP_NAME = "mrtracker"
DB_NAME = "time.db"
CONFIG_FILE = "config.json"
BASE_CONFIG_FILE = "default_config.json"
DB_VERSION = 2

ROOT_PKG_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = user_data_dir(APP_NAME)
CONFIG_DIR = user_config_dir(APP_NAME)
SQL_DIR = os.path.join(ROOT_PKG_DIR, "sql")

DB_PATH = os.path.join(DATA_DIR, DB_NAME)
CONFIG_PATH = os.path.join(CONFIG_DIR, CONFIG_FILE)
BASE_CONFIG_PATH = os.path.join(ROOT_PKG_DIR, BASE_CONFIG_FILE)


def generate_backup_name() -> str:
    return f"mrtracker_{datetime.now().strftime('%Y-%m-%d_%H:%M:%S')}.backup"


def generate_csv_name() -> str:
    return f"mrtracker_{datetime.now().strftime('%Y-%m-%d_%H:%M:%S')}.csv"


def create_dirs(path: str) -> None:
    if not os.path.exists(path):
        create_dirs(os.path.dirname(path))
    try:
        os.mkdir(path)
    except FileExistsError:
        pass


class Config:
    def __init__(self) -> None:
        self.load_config()

    def copy_base_config(self) -> None:
        create_dirs(CONFIG_DIR)
        shutil.copy(BASE_CONFIG_PATH, CONFIG_PATH)

    def load_config(self) -> None:
        if not os.path.exists(CONFIG_PATH):
            self.copy_base_config()
        else:
            self.merge_configs()
        self.read_config()

    def merge_configs(self) -> None:
        with open(CONFIG_PATH, "r") as f1:
            user = json.load(f1)
        with open(BASE_CONFIG_PATH, "r") as f:
            base = json.load(f)

        for section in base:
            if section == "help" or section not in user:
                continue
            for key in base[section]:
                if key in user[section]:
                    base[section][key] = user[section][key]

        with open(CONFIG_PATH, "w") as f1:
            json.dump(base, f1, indent=4)

    def read_config(self) -> None:
        with open(CONFIG_PATH, "r") as f:
            cfg = json.load(f)
            self.app_keys = cfg["app_keys"]
            self.tasklist_keys = cfg["tasklist_keys"]
            self.stats_keys = cfg["stats_keys"]
            self.styles = cfg["styles"]


create_dirs(DATA_DIR)
config = Config()



from textual.reactive import watch
from textual.views._grid_view import GridView

from ..db import Database
from ..config import config
from ..events import DbUpdate
from ..stopwatch import sec_to_str
from ..widgets.current_task import CurrentTask
from ..widgets.header import MyHeader
from ..widgets.in_app_logger import ialogger
from ..widgets.simple_scrollview import SimpleScrollView
from ..widgets.tasklist import TaskList
from ..widgets.timer import Timer


db = Database()


class MainView(GridView):
    def __init__(self, name: str | None = "MainView") -> None:
        super().__init__(name=name)
        self._init_widgets()
        self._make_grid()

    def _init_widgets(self) -> None:
        self.header = MyHeader()
        self.current_task = CurrentTask()
        self.timer = Timer()
        self.ialogger = ialogger
        self.tasklist = TaskList()
        self.t_scroll = SimpleScrollView(self.tasklist)

    def _make_grid(self) -> None:
        self.grid.add_column("left", fraction=1)
        self.grid.add_column("right", min_size=50, fraction=3)
        self.grid.add_row("header", fraction=1, size=1)
        self.grid.add_row("r1", fraction=1, size=3)
        self.grid.add_row("r2", fraction=1, size=3)
        self.grid.add_row("r3", fraction=1)

    async def on_focus(self) -> None:
        await self.tasklist.focus()

    def on_mount(self) -> None:
        self._place_widgets()
        watch(self.tasklist, "current_task", self.start_task)

    def _place_widgets(self) -> None:
        self.grid.add_areas(
            header="left-start|right-end,header",
            current_task="left,r1",
            focus="left,r2",
            logger="left,r3",
            timelist="right,header-end|r3-end",
        )
        self.grid.place(
            header=self.header,
            current_task=self.current_task,
            focus=self.timer,
            logger=self.ialogger,
            timelist=self.t_scroll,
        )

    async def start_task(self, current_task) -> None:
        if not current_task:
            self.current_task.clear_content()
        else:
            self.set_current_task(current_task)
            self.start_session()

    def set_current_task(self, current_task) -> None:
        self.current_task._content = current_task.title

    def start_session(self) -> None:
        self.timer.start()
        ialogger.update("Running")

    async def save_data(self) -> None:
        self.timer.stop()
        if self.timer.saved_time.seconds and self.tasklist.current_task:
            if self.timer.end_time.date() > self.timer.start_time.date():
                self._split_session_and_save()
            else:
                self._save_session()
            await self.app.post_message_from_child(DbUpdate(self))
            hl = config.styles["LOGGER_HIGHLIGHT"]
            ialogger.update(
                "[b]Session saved[/]\n"
                f"[{hl}]{self.current_task.content}[/] - "
                f"{sec_to_str(self.timer.saved_time.seconds)}"
            )
        self.tasklist.current_task = None
        self.timer.restart()

    def _split_session_and_save(self) -> None:
        end_of_first_day = self.timer.start_time.replace(
            hour=23, minute=59, second=59
        )
        duration1 = (end_of_first_day - self.timer.start_time).seconds + 1
        db.add_session(
            self.tasklist.current_task.id,
            self.timer.start_time.strftime("%Y-%m-%d"),
            self.timer.start_time.strftime("%H:%M:%S"),
            end_of_first_day.strftime("%H:%M:%S"),
        )
        self.tasklist.add_time(duration1)

        start_of_second_day = self.timer.end_time.replace(
            hour=0, minute=0, second=0
        )
        duration2 = (self.timer.end_time - start_of_second_day).seconds
        db.add_session(
            self.tasklist.current_task.id,
            self.timer.end_time.strftime("%Y-%m-%d"),
            start_of_second_day.strftime("%H:%M:%S"),
            self.timer.end_time.strftime("%H:%M:%S"),
        )
        self.tasklist.add_time(duration2)

    def _save_session(self) -> None:
        db.add_session(
            self.tasklist.current_task.id,
            self.timer.start_time.strftime("%Y-%m-%d"),
            self.timer.start_time.strftime("%H:%M:%S"),
            self.timer.end_time.strftime("%H:%M:%S"),
        )
        self.tasklist.add_time(self.timer.saved_time.seconds)

    def discard_session(self) -> None:
        if self.timer._working:
            self.timer.restart()
            self.tasklist.current_task = None
            ialogger.update("Session discarded. Timer reset.")



from datetime import datetime

from rich.console import RenderableType
from rich.table import Table
from textual.reactive import Reactive, watch
from textual.widget import Widget

from ..config import config


class MyHeader(Widget):
    def __init__(self) -> None:
        super().__init__()

    clock: Reactive[bool] = Reactive(True)
    title: Reactive[str] = Reactive("")
    sub_title: Reactive[str] = Reactive("")

    @property
    def full_title(self) -> str:
        return (
            f"{self.title} - {self.sub_title}"
            if self.sub_title
            else self.title
        )

    def get_clock(self) -> str:
        return datetime.now().time().strftime("%X")

    def render(self) -> RenderableType:
        header_table = Table.grid(padding=(0, 1), pad_edge=True, expand=True)
        header_table.row_styles = [config.styles["HEADER"]]
        header_table.add_column(justify="left", ratio=0, width=8)
        header_table.add_column("title", justify="center", ratio=1)
        header_table.add_column("clock", justify="right", width=8)
        header_table.add_row(
            "⏱️", self.full_title, self.get_clock() if self.clock else ""
        )
        return header_table

    async def on_mount(self) -> None:
        self.set_interval(1.0, callback=self.refresh)

        async def set_title(title: str) -> None:
            self.title = title

        async def set_sub_title(sub_title: str) -> None:
            self.sub_title = sub_title

        watch(self.app, "title", set_title)
        watch(self.app, "sub_title", set_sub_title)



from rich.align import Align
from rich.console import RenderableType
from rich.panel import Panel
from textual.reactive import Reactive
from textual.widget import Widget

from ..config import config


class CurrentTask(Widget, can_focus=False):

    _content: Reactive[str] = Reactive("")
    _placeholder: Reactive[str] = Reactive("Select task")

    def __init__(self, name: str | None = "CurrentTask") -> None:
        super().__init__(name=name)

    def clear_content(self) -> None:
        self._content = ""

    @property
    def content(self) -> str:
        return self._content

    @content.setter
    def content(self, new_content: str) -> None:
        self._content = new_content

    def render(self) -> RenderableType:
        if self._content:
            rendr = Align.center(self._content)
            border_style = config.styles["CURRENT_TASK_GOOD_BORDER"]
        else:
            rendr = Align.center(self._placeholder)
            border_style = config.styles["CURRENT_TASK_BAD_BORDER"]

        panel = Panel(
            rendr,
            title="Working on",
            style=config.styles["CURRENT_TASK_TEXT"],
            border_style=border_style,
            height=3,
        )
        return panel



from argparse import ArgumentParser

from rich import print

from . import backup
from .config import generate_backup_name, generate_csv_name


def backup_handler(args) -> None:
    try:
        location = backup.create_backup(args.path)
    except FileNotFoundError as e:
        print(e)
    except NotADirectoryError as e:
        print(e)
    else:
        print(f"Backup created at [yellow]{location}[/]")


def restore_handler(args) -> None:
    try:
        backup.restore_data(args.filename)
    except FileNotFoundError as e:
        print(e)
    else:
        print("[green]Data has been restored")


def csv_handler(args) -> None:
    try:
        location = backup.create_csv(args.path)
    except IsADirectoryError as e:
        print(e)
    except NotADirectoryError as e:
        print(e)
    else:
        print(f"csv vile created at [yellow]{location}[/]")


parser = ArgumentParser(
    prog="mrtracker",
    description="mrtracker - a TUI time tracker.\n"
    + "https://github.com/markbragin/mrtracker",
)

commands_parser = parser.add_subparsers(title="commands")

backup_parser = commands_parser.add_parser(
    name="backup",
    help="create backup",
)
backup_parser.add_argument(
    "-p",
    "--path",
    dest="path",
    default=generate_backup_name(),
    help=f"specify path. Defaults to <current directory>",
)
backup_parser.set_defaults(func=backup_handler)

restore_parser = commands_parser.add_parser(
    name="restore",
    help="restore db from backup (rewrites all your current data) ",
)
restore_parser.add_argument(
    "filename",
    help="specify path to backup",
)
restore_parser.set_defaults(func=restore_handler)

csv_parser = commands_parser.add_parser(
    name="csv",
    help="export data to csv",
)
csv_parser.add_argument(
    "-p",
    "--path",
    dest="path",
    default=generate_csv_name(),
    help=f"specify path. Defaults to <current directory>",
)
csv_parser.set_defaults(func=csv_handler)



import sys


def main():
    if len(sys.argv) < 2:
        from .app import TimeTracker
        TimeTracker.run(title="MRTracker")
    else:
        from .arg_parser import parser
        args = parser.parse_args()
        args.func(args)


if __name__ == "__main__":
    main()



from textual.reactive import events

from .text_input import TextInput


class Entry(TextInput):

    _id: int
    _project_id: int | None
    _title: str
    _time: int
    _tag: str | None

    def __init__(self, data: tuple) -> None:
        super().__init__()
        self._id = data[0]
        self._project_id = data[1]
        self._title = data[2]
        self._time = data[3] if data[3] else 0
        self._tag = data[4]

    def on_key(self, event: events.Key) -> None:
        super().on_key(event)

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, new_id) -> None:
        self._id = new_id

    @property
    def project_id(self) -> int | None:
        return self._project_id

    @project_id.setter
    def project_id(self, new_project_id: int) -> None:
        self._project_id = new_project_id

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, new_title: str) -> None:
        self._title = new_title

    @property
    def time(self) -> int:
        return self._time

    @time.setter
    def time(self, new_time: int) -> None:
        self._time = new_time if new_time >= 0 else self._time

    @property
    def type(self) -> str:
        return "task" if self._project_id else "project"

    @property
    def tag(self) -> str | None:
        return self._tag

    @tag.setter
    def tag(self, new_tag: str) -> None:
        self._tag = new_tag if new_tag else None


def generate_entry(entry_id: int, project_id: int | None) -> Entry:
    return Entry((entry_id, project_id, "", 0, None))



from datetime import datetime, timedelta


class Stopwatch:

    _on: bool = False
    _start_t: datetime = datetime.now()
    _saved_time = timedelta()

    @property
    def on(self) -> bool:
        return self._on

    def start(self) -> None:
        self._on = True
        self._start_t = datetime.now()

    def stop(self) -> None:
        self._on = False
        self._saved_time = datetime.now() - self._start_t

    def restart(self) -> None:
        self._on = False
        self._saved_time = timedelta()

    @property
    def elapsed_time(self) -> timedelta:
        if self._on:
            return datetime.now() - self._start_t
        else:
            return timedelta()

    @property
    def saved_time(self) -> timedelta:
        return self._saved_time

    @property
    def start_time(self) -> datetime:
        return self._start_t


def sec_to_str(seconds) -> str:
    if not isinstance(seconds, int):
        return "00:00:00"
    time = f"{seconds % 3600 // 60}:".zfill(3) + f"{seconds % 60}".zfill(2)
    if seconds // 3600:
        time = f"{seconds // 3600}:{time}"
    return time



from sqlite3 import IntegrityError

from rich.console import RenderableType
from rich.padding import PaddingDimensions
from rich.panel import Panel
from rich.table import Table
from rich.text import Text, TextType
from textual.reactive import Reactive, events
from textual.widgets import NodeID, TreeNode

from ..db import Database
from ..config import config
from ..events import DbUpdate
from ..mode import Action, Mode
from ..stopwatch import sec_to_str
from .entry import Entry, generate_entry
from .in_app_logger import ialogger
from .nested_list import NestedList

db = Database()


HL = config.styles["LOGGER_HIGHLIGHT"]

cyrillic_layout = dict(
    zip(
        map(
            ord,
            "йцукенгшщзхъфывапролджэячсмитьбю.ё"
            "ЙЦУКЕНГШЩЗХЪФЫВАПРОЛДЖЭЯЧСМИТЬБЮ,Ё",
        ),
        "qwertyuiop[]asdfghjkl;'zxcvbnm,./`"
        'QWERTYUIOP{}ASDFGHJKL:"ZXCVBNM<>?~',
    )
)


class TaskList(NestedList):

    current_task: Reactive[Entry | None] = Reactive(None)
    _selected: Reactive[NodeID | None] = Reactive(None)
    _mode: Reactive[Mode] = Reactive(Mode.NORMAL)
    _action: Action | None = None
    _mem: NodeID = NodeID(0)
    _show_tags: Reactive[bool] = Reactive(False)
    _list_style: Reactive[int] = Reactive(config.styles["DEFAULT_FORMAT"])

    def __init__(
        self,
        label: TextType = "root",
        data=None,
        name: str | None = None,
        padding: PaddingDimensions = 0,
    ) -> None:
        super().__init__(label=label, data=data, name=name, padding=padding)
        self._tree.hide_root = True

    async def on_mount(self) -> None:
        await self.collect_data()

    async def collect_data(self) -> None:
        self._projects = db.fetch_projects()
        self._tasks = db.fetch_tasks()
        await self.add_header()
        await self._build_tree()
        await self.root.expand()
        self.sum_projects_time()

    async def add_header(self) -> None:
        self.id = -2
        await self.add(NodeID(0), "header", None)
        self.id = 0

    async def _build_tree(self) -> None:
        for row in self._projects:
            self.id = row[0] - 1
            await self.add(NodeID(0), row[1], Entry(row))

        for row in self._tasks:
            await self.add(row[1], row[2], Entry(row))

    def sum_projects_time(self) -> None:
        for project in self.root.children[1:]:
            project.data.time = sum(
                task.data.time for task in project.children
            )

    async def on_focus(self) -> None:
        self._mode = Mode.NORMAL

    async def on_key(self, event: events.Key) -> None:
        if self._mode is Mode.NORMAL:
            await self._handle_keypress_in_normal_mode(event)
        elif self._mode is Mode.INSERT:
            await self._handle_keypress_in_insert_mode(event)

    async def _handle_keypress_in_normal_mode(self, event: events.Key) -> None:
        key = event.key.translate(cyrillic_layout)
        if event.key == "escape":
            await self._cancel()
        elif event.key == "enter":
            if self._action == Action.MOVE:
                await self._handle_moving_entry()
            else:
                self._handle_starting_task()
        elif key in [config.tasklist_keys["go_down"], "down"]:
            await self.go_down()
        elif key in [config.tasklist_keys["go_up"], "up"]:
            await self.go_up()
        elif key == config.tasklist_keys["add_task"]:
            await self.add_task()
        elif key == config.tasklist_keys["new_project"]:
            await self.add_project()
        elif key == config.tasklist_keys["rename_entry"]:
            self.rename_entry()
        elif key == config.tasklist_keys["delete_entry"]:
            self.delete_entry()
        elif key == config.tasklist_keys["reset_entry_time"]:
            self.reset_entry_time()
        elif key == config.tasklist_keys["move_entry"]:
            self.move_entry()
        elif key == config.tasklist_keys["change_tag"]:
            self.change_tag()
        elif key == config.tasklist_keys["show_tags"]:
            self.toggle_tags()
        elif key == config.tasklist_keys["toggle_project"]:
            await self.toggle_project()
        elif key == config.tasklist_keys["toggle_all_projects"]:
            await self.toggle_all_projects()
        elif key in ["1", "2", "3"]:
            self._list_style = int(key)

    async def _cancel(self) -> None:
        if not self._action:
            return
        elif self._action is Action.ADD:
            await self.remove_node()
        ialogger.update("Canceled")
        self._action = None
        self._selected = None
        self._mode = Mode.NORMAL
        self.cursor = self._mem

    async def _handle_moving_entry(self) -> None:
        selected = self.nodes[self._selected]
        curr = self.nodes[self.cursor]
        if selected.parent is curr.parent:
            self._swap_entries()
        elif selected.data.type == "task" and curr.data.type == "project":
            self._change_project()
        else:
            ialogger.update(
                f"Can't swap [{HL}]{selected.data.title}[/] ⮀ "
                f"[{HL}]{curr.data.title}[/]",
                error=True,
            )
            return
        self.cursor = self._selected if self._selected else self.cursor
        self._selected = None
        self._action = None
        await self.app.post_message_from_child(DbUpdate(self))
        ialogger.update("[b]DONE[/]")

    def _swap_entries(self) -> None:
        one = self.nodes[self._selected]
        two = self.nodes[self.cursor]
        self._swap_entry_ids(one.data, two.data)
        self._swap_trees(one, two)
        self._swap_nodes(one, two)
        if one.data.type == "task":
            db.swap_tasks(one.data.id, two.data.id)
        else:
            db.swap_projects(one.data.id, two.data.id)

    def _swap_entry_ids(self, one: Entry, two: Entry) -> None:
        one.id, two.id = two.id, one.id

    def _swap_nodes(self, one: TreeNode, two: TreeNode) -> None:
        children = one.parent.children
        idx1 = children.index(one)
        idx2 = children.index(two)
        children[idx1], children[idx2] = children[idx2], children[idx1]

    def _swap_trees(self, one: TreeNode, two: TreeNode) -> None:
        children = one.parent.tree.children
        idx1 = children.index(one.tree)
        idx2 = children.index(two.tree)
        children[idx1], children[idx2] = children[idx2], children[idx1]

    def _change_project(self) -> None:
        selected_node = self.nodes[self._selected]
        curr_node = self.nodes[self.cursor]
        self._move_to_new_parent(selected_node, curr_node)
        selected_node.data.project_id = curr_node.data.id
        db.change_project(selected_node.data.id, curr_node.data.id)
        self.sum_projects_time()

    def _move_to_new_parent(self, node: TreeNode, new_par: TreeNode) -> None:
        new_par.tree.children.append(node.tree)
        node.parent.tree.children.remove(node.tree)

        new_par.children.append(node)
        node.parent.children.remove(node)

        node.parent = new_par
        node.data.parent_id = new_par.data.id

    def _handle_starting_task(self) -> None:
        if self.current_task:
            ialogger.update("[b]Timer is already running[/]", error=True)
        else:
            entry = self.nodes[self.cursor].data
            if entry.type == "project":
                ialogger.update("Select task, not project", error=True)
            else:
                self.current_task = entry

    async def go_down(self) -> None:
        await self.cursor_down()

    async def go_up(self) -> None:
        if self.nodes[self.cursor].previous_node is not self.root.children[0]:
            await self.cursor_up()

    async def add_task(self) -> None:
        if not self._valid_cursor():
            return
        node = self.nodes[self.cursor]
        self._mem = self.cursor
        if node.data.type == "project":
            await self.add_child(
                "task",
                generate_entry(db.get_next_task_id(), node.data.id),
            )
        else:
            await self.add_sibling(
                "task",
                generate_entry(db.get_next_task_id(), node.parent.data.id),
            )
        self._action = Action.ADD
        self._mode = Mode.INSERT
        node.data.clear_content()

    async def add_project(self) -> None:
        self._mem = self.cursor
        await self.add_root_child(
            "project",
            generate_entry(db.get_next_project_id(), None),
        )
        self._action = Action.ADD
        self._mode = Mode.INSERT
        self.nodes[self.cursor].data.clear_content()

    def rename_entry(self) -> None:
        if not self._valid_cursor():
            return
        self._mem = self.cursor
        self._action = Action.RENAME
        self._mode = Mode.INSERT
        entry = self.nodes[self.cursor].data
        entry.content = entry.title
        ialogger.update(f"[b]Rename[/]\nType new name")

    def delete_entry(self) -> None:
        if not self._valid_cursor():
            return
        self._mem = self.cursor
        self._action = Action.DELETE
        self._mode = Mode.INSERT
        entry = self.nodes[self.cursor].data
        entry.clear_content()
        ialogger.update(
            f"Type [{HL}]'delete'[/] to delete [{HL}]{entry.title}[/]"
        )

    def reset_entry_time(self) -> None:
        if not self._valid_cursor():
            return
        self._mem = self.cursor
        entry = self.nodes[self.cursor].data
        self._action = Action.RESET
        self._mode = Mode.INSERT
        entry.clear_content()
        ialogger.update(
            f"Type [{HL}]'reset'[/] to reset [{HL}]{entry.title}[/] time"
        )

    def move_entry(self) -> None:
        if not self._valid_cursor():
            return
        self._mem = self.cursor
        self._action = Action.MOVE
        self._selected = self.nodes[self.cursor].id
        ialogger.update("Select target")

    def change_tag(self) -> None:
        if not self._valid_cursor():
            return
        self._mem = self.cursor
        self._action = Action.ADD_TAG
        self._mode = Mode.INSERT
        self.nodes[self.cursor].data.clear_content()
        ialogger.update("Type new tag")

    def toggle_tags(self) -> None:
        self._show_tags = not self._show_tags

    def _valid_cursor(self) -> bool:
        return self.cursor not in [self.root.id, self.root.children[0].id]

    async def toggle_project(self) -> None:
        if self.nodes[self.cursor].parent is not self.root:
            self._cur_to_parent()
        await self.nodes[self.cursor].toggle()

    async def toggle_all_projects(self) -> None:
        if not self._valid_cursor():
            return
        for project in self.root.children:
            await project.expand(not self.root.children[-1].expanded)
        if self.nodes[self.cursor].data.type == "task":
            self._cur_to_parent()

    async def _handle_keypress_in_insert_mode(self, event: events.Key) -> None:
        self.nodes[self.cursor].data.on_key(event)
        if event.key == "escape":
            await self._cancel()
        elif event.key == "enter":
            if self._action is Action.ADD:
                await self._handle_adding_entry()
            elif self._action is Action.RENAME:
                await self._handle_renaming_entry()
            elif self._action is Action.DELETE:
                await self._handle_deleting_entry()
            elif self._action is Action.RESET:
                await self._handle_resetting_task_time()
            elif self._action is Action.ADD_TAG:
                await self._handle_changing_tag()
            self._action = None
            self._mode = Mode.NORMAL
        event.stop()
        self.refresh()

    async def _handle_adding_entry(self) -> None:
        entry = self.nodes[self.cursor].data
        if entry.content == entry.title == "":
            await self._cancel()
            return

        entry.title = entry.content
        if entry.type == "project":
            try:
                db.add_project(entry.title)
            except IntegrityError:
                ialogger.update(
                    "[red]ERROR[/]. "
                    + f"Project with name {entry.title} already exists",
                    error=True,
                )
                await self.remove_node()

        else:
            db.add_task(entry.title, entry.project_id)

    async def _handle_renaming_entry(self) -> None:
        entry = self.nodes[self.cursor].data
        if not entry.content or entry.title == entry.content:
            await self._cancel()
            return

        entry.title = entry.content
        if entry.type == "project":
            db.rename_project(entry.id, entry.title)
        else:
            db.rename_task(entry.id, entry.title)
        ialogger.update("[b]DONE[/]")

    async def _handle_deleting_entry(self) -> None:
        node = self.nodes[self.cursor]
        entry = node.data
        if entry.content != "delete":
            await self._cancel()
            return

        if entry.type == "project":
            db.delete_project(entry.id)
        else:
            db.delete_task(entry.id)
        await self.remove_node()
        await self.go_down()
        self.sum_projects_time()
        await self.app.post_message_from_child(DbUpdate(self))
        ialogger.update("[b]DONE[/]")

    async def _handle_resetting_task_time(self) -> None:
        node = self.nodes[self.cursor]
        entry = node.data
        if entry.content != "reset":
            await self._cancel()
            return

        if entry.type == "project":
            db.delete_sessions_by_task_ids(
                [child.data.id for child in node.children]
            )
            for t in node.children:
                t.data.time = 0
        else:
            db.delete_sessions_by_task_ids([entry.id])
            entry.time = 0
        self.sum_projects_time()
        await self.app.post_message_from_child(DbUpdate(self))
        ialogger.update("[b]DONE[/]")

    async def _handle_changing_tag(self) -> None:
        node = self.nodes[self.cursor]
        entry = node.data
        new_tag = entry.content if entry.content else None
        task_ids = []
        if entry.type == "project":
            for task_node in node.children:
                task_node.data.tag = new_tag
                task_ids.append(task_node.data.id)
        else:
            entry.tag = new_tag
            task_ids.append(entry.id)
        db.update_tags(task_ids, new_tag)
        await self.app.post_message_from_child(DbUpdate(self))
        ialogger.update("[b]DONE[/]")

    def add_time(self, time: int) -> None:
        self.current_task.time += time
        self.sum_projects_time()

    def render(self) -> RenderableType:
        return Panel(self._tree, border_style=config.styles["TASKLIST_BORDER"])

    def render_node(self, node: TreeNode) -> RenderableType:
        if node is self.root:
            return self._render_root()
        elif node is self.root.children[0]:
            return self._render_header()
        else:
            rendr = self._render_entry(node)
        if node.id == self.cursor:
            rendr.row_styles = [config.styles["HIGHLIGHTED_ENTRY"]]
        elif node.id == self._selected:
            rendr.row_styles = [config.styles["SELECTED_ENTRY"]]
        elif node.data.type == "project":
            rendr.row_styles = [config.styles["PROJECT_STYLE"]]
        elif node.data.type == "task":
            rendr.row_styles = [config.styles["TASK_STYLE"]]
        return rendr

    def _render_root(self) -> RenderableType:
        return self.root.label

    def _render_header(self) -> RenderableType:
        style = config.styles["TASKLIST_HEADER"]
        if self._list_style == 3:
            header = Table.grid(expand=True)
            header.add_row("Projects", Text("Time", justify="right"))
            header.row_styles = [style]
        else:
            header = f"[{style}]Projects"
        return header

    def _render_entry(self, node: TreeNode) -> Table:
        grid = Table.grid(expand=True)
        name = self._render_name(node)
        time = self._render_time(node) if node.data.title else Text("")
        if self._list_style == 1:
            grid.add_row(Text.assemble(name, time))
        elif self._list_style == 2:
            grid.add_row(Text.assemble(time, name))
        else:
            grid.add_row(name, time)
        return grid

    def _render_name(self, node: TreeNode) -> Text:
        cursor = self.cursor == node.id and self._mode is Mode.INSERT
        name = node.data.title
        if node.data.type == "project":
            if node.expanded:
                name = f"⇣ {name}"
            else:
                name = f"→ {name}"
                if node.children:
                    name = f"{name} [{len(node.children)}]"
        elif self._show_tags:
            name = f"{name} #{node.data.tag if node.data.tag else ''}"
        name = Text(name)
        if cursor:
            name = node.data._render_with_cursor()
        return name

    def _render_time(self, node: TreeNode) -> Text:
        justify = "right" if self._list_style == 3 else "left"
        return Text(
            f" {sec_to_str(node.data.time)} ",
            justify=justify,
        )



from rich.console import RenderableType
from textual.message import Message
from textual.widgets import ScrollView


class SimpleScrollView(ScrollView):
    async def update(self, content: RenderableType) -> None:
        await super().update(content)

    async def handle_window_change(self, message: Message) -> None:
        message.stop()

        virtual_width, virtual_height = self.window.virtual_size
        width, height = self.size

        self.x = self.validate_x(self.x)
        self.y = self.validate_y(self.y)

        self.hscroll.virtual_size = virtual_width
        self.hscroll.window_size = width
        self.vscroll.virtual_size = virtual_height
        self.vscroll.window_size = height

    async def key_down(self) -> None:
        pass

    async def key_up(self) -> None:
        pass



from enum import Enum, auto


class Mode(Enum):
    NORMAL = auto()
    INSERT = auto()


class Action(Enum):
    ADD = auto()
    RENAME = auto()
    DELETE = auto()
    RESET = auto()
    MOVE = auto()
    ADD_TAG = auto()



from rich.align import Align
from rich.console import RenderableType
from rich.panel import Panel
from textual.reactive import Reactive
from textual.widget import Widget

from ..__init__ import __version__
from ..config import config


class Logger(Widget, can_focus=False):

    _content: Reactive[str] = Reactive("")
    _error: Reactive[bool] = Reactive(False)

    def __init__(
        self,
        name: str | None = "Logger",
    ) -> None:
        super().__init__(name=name)
        self._content = (
            f"MRTracker v{__version__}\n"
            + f"[{config.styles['LOGGER_HIGHLIGHT']}]"
            + f"{config.app_keys['show_help']}[/] - show help"
        )

    @property
    def content(self) -> str:
        return self._content

    @content.setter
    def content(self, new_content) -> None:
        self._content = new_content

    @property
    def error(self) -> bool:
        return self._error

    @error.setter
    def error(self, upd) -> None:
        self._error = upd

    def update(self, msg: str, error: bool = False):
        self._error = error
        self._content = msg

    def render(self) -> RenderableType:
        if self._error:
            border_style = config.styles["LOGGER_ERROR_BORDER"]
        else:
            border_style = config.styles["LOGGER_BORDER"]

        self.panel = Panel(
            Align.center(self.content, vertical="middle", style="white"),
            border_style=border_style,
            expand=True,
        )
        return self.panel


ialogger = Logger()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mrtracker', 'mrtracker.views', 'mrtracker.widgets']

package_data = \
{'': ['*'], 'mrtracker': ['sql/*']}

install_requires = \
['platformdirs>=2.5.2,<3.0.0', 'textual>=0.1.18,<0.2.0']

entry_points = \
{'console_scripts': ['mrtracker = mrtracker.__main__:main']}

setup_kwargs = {
    'name': 'mrtracker',
    'version': '0.4.2',
    'description': 'A TUI time tracker',
    'long_description': None,
    'author': 'Mark Bragin',
    'author_email': 'm4rk.brag1n@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/markbragin/mrtracker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from rich.align import Align
from rich.table import Table
from textual.app import events
from textual.views._dock_view import DockView

from ..config import CONFIG_PATH, config
from ..widgets.simple_scrollview import SimpleScrollView


class HelpView(DockView):
    def __init__(self, name: str | None = "HelpView") -> None:
        super().__init__(name=name)
        self.create_table()
        self.scrll = SimpleScrollView(Align(self.table, "center"))

    async def on_focus(self) -> None:
        await self.scrll.focus()

    async def on_mount(self) -> None:
        await self.dock(self.scrll)

    def create_table(self):
        self.table = Table(show_header=False, box=None)
        self.table.title_style = ""
        self._insert_title()
        self._insert_app_keys()
        self._insert_tasklist_keys()
        self._insert_stats_keys()

    def _insert_title(self) -> None:
        self.table.title = (
            "[yellow bold]Help\n\n[/]"
            + "To change bindings and styles edit\n"
            + f"[dim]{CONFIG_PATH}[/]"
        )

    def _insert_app_keys(self) -> None:
        self.table.add_row("[magenta]App keys:")
        for descrip, key in config.app_keys.items():
            self.table.add_row(
                Align(key, "right", style="blue"), " ".join(descrip.split("_"))
            )
        self.table.add_row(end_section=True)

    def _insert_tasklist_keys(self) -> None:
        self.table.add_row("[magenta]Tasklist keys:")
        for descrip, key in config.tasklist_keys.items():
            self.table.add_row(
                Align(key, "right", style="blue"),
                " ".join(descrip.split("_")),
            )
        self.table.add_row(
            Align("1, 2, 3", "right", style="blue"),
            "change format",
        )
        self.table.add_row(end_section=True)

    def _insert_stats_keys(self) -> None:
        self.table.add_row("[magenta]Stats keys:")
        for descrip, key in config.stats_keys.items():
            self.table.add_row(
                Align(key, "right", style="blue"),
                " ".join(descrip.split("_")),
            )

    async def on_key(self, event: events.Key) -> None:
        if event.key in [config.tasklist_keys["go_down"], "down"]:
            self.scrll.page_down()
        elif event.key in [config.tasklist_keys["go_up"], "up"]:
            self.scrll.page_up()



from typing import Literal

from rich.panel import Panel
from rich.table import Table
from textual import events
from textual.views._grid_view import GridView

from ..db import Database
from ..config import config
from ..stopwatch import sec_to_str
from ..widgets.simple_scrollview import SimpleScrollView


TimeInterval = Literal["today", "week", "month"]

SSS = config.styles["STATS_SUBHEADERS_STYLE"]
SPS = config.styles["STATS_PROJECTS_STYLE"]

db = Database()


class StatsView(GridView):
    _show_project: bool = False
    _focused: SimpleScrollView | None = None
    _need_update: bool = False

    today: SimpleScrollView
    week: SimpleScrollView
    month: SimpleScrollView

    projects: dict[TimeInterval, list[tuple]] = dict()
    tasks: dict[TimeInterval, list[tuple]] = dict()
    tags: dict[TimeInterval, list[tuple]] = dict()

    def __init__(self, name: str | None = "StatsView") -> None:
        super().__init__(name=name)
        self._make_view_grid()
        self._collect_data()
        self._init_widgets()

    def _make_view_grid(self) -> None:
        self.grid.add_column("left", fraction=1)
        self.grid.add_column("center", fraction=1)
        self.grid.add_column("right", fraction=1)
        self.grid.add_row("row")

    def _collect_data(self) -> None:
        self.projects["today"] = db.fetch_projects_today()
        self.projects["week"] = db.fetch_projects_week()
        self.projects["month"] = db.fetch_projects_month()
        self.tasks["today"] = db.fetch_tasks_today()
        self.tasks["week"] = db.fetch_tasks_week()
        self.tasks["month"] = db.fetch_tasks_month()
        self.tags["today"] = db.fetch_tags_today()
        self.tags["week"] = db.fetch_tags_week()
        self.tags["month"] = db.fetch_tags_month()

    def _init_widgets(self) -> None:
        self.today = SimpleScrollView(self._get_today_section())
        self.week = SimpleScrollView(self._get_week_section())
        self.month = SimpleScrollView(self._get_month_section())

    def _get_today_section(self) -> Panel:
        grid = self._get_section_grid()
        self._fill_grid(grid, "today")
        return Panel(
            grid,
            title="Today",
            border_style=config.styles["STATS_TODAY_BORDER_STYLE"],
        )

    def _get_week_section(self) -> Panel:
        grid = self._get_section_grid()
        self._fill_grid(grid, "week")
        return Panel(
            grid,
            title="Last 7 days",
            border_style=config.styles["STATS_WEEK_BORDER_STYLE"],
        )

    def _get_month_section(self) -> Panel:
        grid = self._get_section_grid()
        self._fill_grid(grid, "month")
        return Panel(
            grid,
            title="Last 30 days",
            border_style=config.styles["STATS_MONTH_BORDER_STYLE"],
        )

    def _get_section_grid(self) -> Table:
        grid = Table.grid(expand=True, padding=(0, 1))
        grid.add_column("Task", justify="left", overflow="fold")
        grid.add_column("Time", justify="right", no_wrap=True)
        grid.row_styles = [config.styles["STATS_NORMAL_TEXT"]]
        return grid

    def _fill_grid(self, grid: Table, ti: TimeInterval) -> None:
        self._insert_projects(grid, ti)
        self._insert_tasks(grid, ti)
        self._insert_tags(grid, ti)

    def _insert_projects(self, grid: Table, ti: TimeInterval) -> None:
        grid.add_row(f"[{SSS}]Projects:")
        for row in self.projects[ti]:
            grid.add_row(row[0], sec_to_str(row[1]))
        grid.add_row(end_section=True)

    def _insert_tasks(self, grid: Table, ti: TimeInterval) -> None:
        grid.add_row(f"[{SSS}]Tasks:")
        for row in self.tasks[ti]:
            task = (
                f"{row[0]} -> [{SPS}]{row[2]}[/]"
                if self._show_project
                else row[0]
            )
            grid.add_row(task, sec_to_str(row[1]))
        grid.add_row(end_section=True)

    def _insert_tags(self, grid: Table, ti: TimeInterval) -> None:
        grid.add_row(f"[{SSS}]Tags:")
        for row in self.tags[ti]:
            grid.add_row(row[0], sec_to_str(row[1]))

    async def on_mount(self) -> None:
        self._place_widgets()

    def _place_widgets(self) -> None:
        self.grid.add_areas(
            left="left,row",
            center="center,row",
            right="right,row",
        )
        self.grid.place(
            left=self.today,
            center=self.week,
            right=self.month,
        )

    async def _update(self) -> None:
        self._collect_data()
        await self._rerender()

    async def _rerender(self) -> None:
        await self.today.update(self._get_today_section())
        await self.week.update(self._get_week_section())
        await self.month.update(self._get_month_section())

    async def on_focus(self) -> None:
        if not self._focused:
            self._focused = self.today
        await self._focused.focus()
        if self._need_update:
            await self._update()
            self._need_update = False

    async def on_key(self, event: events.Key) -> None:
        if event.key == config.stats_keys["toggle_projects_after_task"]:
            self._show_project = not self._show_project
            await self._update()
        elif event.key in ["left", config.stats_keys["left"]]:
            await self._focus_left()
        elif event.key in ["right", config.stats_keys["right"]]:
            await self._focus_right()
        elif event.key in ["down", config.stats_keys["down"]]:
            if self._focused:
                self._focused.page_down()
        elif event.key in ["up", config.stats_keys["up"]]:
            if self._focused:
                self._focused.page_up()

    async def _focus_left(self) -> None:
        if self._focused is None:
            self._focused = self.month
        elif self._focused == self.month:
            self._focused = self.week
        elif self._focused == self.week:
            self._focused = self.today

    async def _focus_right(self) -> None:
        if self._focused is None:
            self._focused = self.today
        elif self._focused == self.today:
            self._focused = self.week
        elif self._focused == self.week:
            self._focused = self.month

    def require_update(self) -> None:
        self._need_update = True



import csv
import os
import shutil

from .config import DB_PATH, generate_backup_name, generate_csv_name
from .db import Database


def create_backup(path: str) -> str:
    """Creates backup and returns its location"""
    if os.path.isdir(path):
        backup_path = os.path.join(path, generate_backup_name())
    elif os.path.isdir(os.path.dirname(path)):
        backup_path = path
    else:
        backup_path = os.path.join(".", path)

    shutil.copy(DB_PATH, backup_path)
    return os.path.abspath(backup_path)


def restore_data(path: str) -> None:
    """Replaces all the data with the data from backup"""
    shutil.copy(path, DB_PATH)


def create_csv(path: str) -> str:
    """Creates csv file and returns its location"""
    if os.path.isdir(path):
        csv_path = os.path.join(path, generate_csv_name())
    elif os.path.isdir(os.path.dirname(path)):
        csv_path = path
    else:
        csv_path = os.path.join(".", path)

    data = Database().fetch_for_csv()
    with open(csv_path, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(
            (
                "project",
                "task",
                "tags",
                "date",
                "start time",
                "end time",
                "duration",
            )
        )
        writer.writerows(data)
    return os.path.abspath(csv_path)



from textual.app import App
from textual.layouts.dock import DockLayout
from textual.reactive import Reactive, events, watch
from textual.view import View
from textual.views._grid_view import GridView

from .config import config
from .views.help_view import HelpView
from .views.main_view import MainView
from .views.stats_view import StatsView
from .widgets.in_app_logger import ialogger


class TimeTracker(App):

    current_view: Reactive[View | None] = Reactive(None)

    async def on_load(self) -> None:
        for action, key in config.app_keys.items():
            await self.bind(key, action)

    async def on_db_update(self) -> None:
        self.stats_v.require_update()

    async def on_mount(self) -> None:
        self.main_v = MainView()
        self.help_v = HelpView()
        self.stats_v = StatsView()
        self.current_view = self.main_v
        await self.main_v.tasklist.post_message(events.Key(self, "down"))
        await self.main_v.tasklist.post_message(events.Key(self, "down"))
        watch(self, "current_view", self.update_view)

    async def update_view(self, view: GridView) -> None:
        self.clear_screen()
        await self.view.dock(view)
        await view.focus()

    def clear_screen(self) -> None:
        if isinstance(self.view.layout, DockLayout):
            self.view.layout.docks.clear()
        self.view.widgets.clear()

    async def action_quit(self) -> None:
        ialogger.update("[i]Saving data...[/]")
        await self.main_v.save_data()
        await self.shutdown()

    async def action_reset_focus(self) -> None:
        self.current_view = self.main_v

    async def action_save_session(self) -> None:
        await self.main_v.save_data()

    def action_discard_session(self) -> None:
        self.main_v.discard_session()

    async def action_show_help(self) -> None:
        self.current_view = self.help_v

    async def action_show_stats(self) -> None:
        self.current_view = self.stats_v



from datetime import datetime, timedelta

from rich.align import Align
from rich.console import RenderableType
from rich.panel import Panel
from textual.reactive import Reactive
from textual.widget import Widget

from ..config import config
from ..stopwatch import Stopwatch, sec_to_str


class Timer(Widget, can_focus=False):

    _working: Reactive[bool] = Reactive(False)

    def __init__(self, name: str | None = "Timer") -> None:
        super().__init__(name=name)
        self.stopwatch = Stopwatch()

    async def on_mount(self):
        self.set_interval(0.5, self.refresh)

    @property
    def elapsed_time(self) -> timedelta:
        return self.stopwatch.elapsed_time

    def get_elapsed_time_str(self) -> str:
        """returns formatted elapsed time [%H:]%M:%S"""
        return sec_to_str(self.stopwatch.elapsed_time.seconds)

    @property
    def saved_time(self) -> timedelta:
        return self.stopwatch.saved_time

    @property
    def start_time(self) -> datetime:
        return self.stopwatch.start_time

    @property
    def end_time(self) -> datetime:
        return self.stopwatch.start_time + timedelta(
            seconds=self.stopwatch.saved_time.seconds
        )

    @property
    def working(self) -> bool:
        return self._working

    def start(self) -> None:
        self.stopwatch.start()

    def stop(self) -> None:
        self.stopwatch.stop()

    def restart(self):
        self.stopwatch.restart()

    def render(self) -> RenderableType:
        self._working = self.stopwatch.on
        self.panel = Panel(
            Align.center(
                self.get_elapsed_time_str(),
                vertical="middle",
                style=config.styles["TIMER_TEXT"],
            ),
            title="for",
            border_style=config.styles["TIMER_BORDER"],
            expand=True,
        )
        return self.panel
